% This code is for the hands-on portion of the 2023 ISB Workshop:
% Power to the Foot! Understanding and Applying Distal Power Calculations
%
% The purpose of this code is to demonstrate the utility of distal shank
% and rearfoot power in prosthetic and able-bodied walking.
%
% The functions used for this code are contained at the end of the script.
% These are called "local functions" and require MATLAB version 2016b or
% higher. If you have an older version of MATLAB, please copy the functions
% into their own function files in order to run this script. 
%
% Written on: July 7, 2023
% Written by: Eric Honert and Kota Takahashi
%__________________________________________________________________________
clear; clc; close all;

%% Person with Amputation Walking
% Note: Sound limb = Right, Amputated limb = Left
dat_amp = load('amputee_R&Lshank.mat');
% Fill the NaN values for COP & Free Moment with 0
dat_amp.LCOP(isnan(dat_amp.LCOP)) = 0;
dat_amp.LFreeMom(isnan(dat_amp.LFreeMom)) = 0;
dat_amp.RCOP(isnan(dat_amp.RCOP)) = 0;
dat_amp.RFreeMom(isnan(dat_amp.RFreeMom)) = 0;
% Find the left and right foot contact events
dat_amp.LHS  = findHeelStrikes(dat_amp.LGRF(:,3),20);
dat_amp.RHS  = findHeelStrikes(dat_amp.RGRF(:,3),20);
% Compute the velocity of the treadmill from the left and right foot moving 
% backwards on the treadmill. 
dat_amp.LTreadmill_Velocity = find_foot_vel_treadmill(dat_amp.LHS,1.25,dat_amp.LFT_CGVel);
dat_amp.RTreadmill_Velocity = find_foot_vel_treadmill(dat_amp.RHS,1.25,dat_amp.RFT_CGVel);
% Compute the Distal Shank power for the prosthesis & sound limb
dat_amp.LDS_power = DIST_SEG_POWER(dat_amp.LSHK_CGVel, dat_amp.LSHK_AngVel, dat_amp.LSHK_CGPos,...
    dat_amp.LCOP, dat_amp.LFreeMom, dat_amp.LGRF, 1, dat_amp.LTreadmill_Velocity);
dat_amp.RDS_power = DIST_SEG_POWER(dat_amp.RSHK_CGVel, dat_amp.RSHK_AngVel, dat_amp.RSHK_CGPos,...
    dat_amp.RCOP, dat_amp.RFreeMom, dat_amp.RGRF, 1, dat_amp.RTreadmill_Velocity);

% Interpolate the necessary data to 101 points
dat_amp.intp_LDS = intp_stride(dat_amp.LHS,dat_amp.LDS_power);
dat_amp.intp_Lank_pow = intp_stride(dat_amp.LHS,dat_amp.LANKPow3DOF);
dat_amp.intp_RDS = intp_stride(dat_amp.RHS,dat_amp.RDS_power);
dat_amp.intp_Rank_pow = intp_stride(dat_amp.RHS,dat_amp.RANKPow3DOF);

% Compare the prosthesis distal shank power to 3DOF ankle power
figure(1)
plot(0:100,mean(dat_amp.intp_LDS,2),'r','LineWidth', 5)
hold on;
plot(0:100,mean(dat_amp.intp_Lank_pow,2),'k','LineWidth', 5)
hold on;
plot(0:100,dat_amp.intp_LDS,'r')
hold on;
plot(0:100,dat_amp.intp_Lank_pow,'k')
legend('Distal Shank Power','3DOF Ankle Power')
xlabel('% Stride')
ylabel('Power (W)')
title('Comparison of Power from a Prosthesis')

% Compare the prosthesis & sound limb distal shank power
figure(2)
plot(0:100,mean(dat_amp.intp_LDS,2),'r','LineWidth', 5)
hold on;
plot(0:100,mean(dat_amp.intp_RDS,2),'b','LineWidth', 5)
hold on;
plot(0:100,dat_amp.intp_LDS,'r')
hold on;
plot(0:100,dat_amp.intp_RDS,'b')
legend('Distal Shank Power: Prosthesis','Distal Shank Power: Sound Limb')
xlabel('% Stride')
ylabel('Power (W)')
title('Comparison of Power between Prosthesis & Sound Limbs')

%% Able-Bodied Walking
dat_ab = load('nonamputee_hindfoot&shank.mat');
% Fill the NaN values for COP & Free Moment with 0
dat_ab.LCOP(isnan(dat_ab.LCOP)) = 0;
dat_ab.LFreeMom(isnan(dat_ab.LFreeMom)) = 0;
% Find the foot contact events
dat_ab.LHS  = findHeelStrikes(dat_ab.LGRF(:,3),20);
% Compute the velocity of the treadmill from the foot moving backward on
% the treadmill. NOTE: Only for walking
dat_ab.Treadmill_Velocity = find_foot_vel_treadmill(dat_ab.LHS,1.25,dat_ab.LHF_CGVel);
% Compute the distal reafoot power
dat_ab.DR_power = DIST_SEG_POWER(dat_ab.LHF_CGVel, dat_ab.LHF_AngVel, dat_ab.LHF_CGPos,...
    dat_ab.LCOP, dat_ab.LFreeMom, dat_ab.LGRF, 1, dat_ab.Treadmill_Velocity);

% Interpolate the data to 101 points
dat_ab.intp_DR = intp_stride(dat_ab.LHS,dat_ab.DR_power);
dat_ab.intp_ank_pow = intp_stride(dat_ab.LHS,dat_ab.LANKPow3DOF);

% Examine the distal rearfoot vs. the 3 DOF ankle power
figure(3)
plot(0:100,mean(dat_ab.intp_DR,2),'r','LineWidth', 5)
hold on;
plot(0:100,mean(dat_ab.intp_ank_pow,2),'k','LineWidth', 5)
hold on;
plot(0:100,dat_ab.intp_DR,'r')
hold on;
plot(0:100,dat_ab.intp_ank_pow,'k')
legend('Distal Rearfoot Power','3DOF Ankle Power')
xlabel('% Stride')
ylabel('Power (W)')
title('Able-Bodied Walking')

%%
%==========================================================================
% Function block
function HS = findHeelStrikes(GRF_vert,thresh)
% This function computes the initial contact indicies (heel strikes) from 
% the vertical ground reaction force. Note: the ground reaction forces have
% been zeroed below the input threshold
%__________________________________________________________________________
% Parameters
% ----------
% GRF_vert = vertical ground reation force
% thresh = threshold for determining heel strike (ex: 20)
%
% Returns
% -------
% HS = heel strike (foot contact) indicies
%__________________________________________________________________________
count = 1; % counter for storing HS indicies
% Index through the vertical GRF signal
for ii = 1:(length(GRF_vert)-1)
	if GRF_vert(ii) == 0 && GRF_vert(ii+1) > thresh
		HS(count) = ii;
		count = count+1;
	end
end
end

function Dist_Seg_Pow = DIST_SEG_POWER(Seg_COM_Vel,Seg_Ang_Vel,Seg_COM_Pos,CenterOfPressure,FreeMoment,GRF,treadmill_yn,foot_vel)
% The funciton computes the distal "segment" power for a segment of 
% interest (for example: if the function inputs are for the calcaneus, this 
% function will output the distal calcaneus power or if the segment of 
% interest is the shank, the function will output the distal shank power)
% during gait. Treadmill and overground walking are accomidated with this
% code by changing the "treadmill_yn" and "foot_vel" inputs.
%__________________________________________________________________________
% Parameters
% ----------
% Seg_COM_Vel = Segment COM Velocity (ex: Calcaneus COM Velocity)
% Seg_Ang_Vel = Segment Angular Velocity (ex: Calcanues Angular Velocity)
% Seg_COM_Pos = Segment COM Position (ex: Calcaneus COM Position)
% CenterOfPressure = Location of the center of pressure in the lab
%   coordinate system
% FreeMoment = Free moment on force platform in the lab coordinate system
% GRF = Ground Reaction Force in the lab coordinate system
% treadmill_yn = 0 if overground, 1 if on the treadmill 
% foot_vel = velocity of the foot on the ground. This can be obtained from
%   the function "find_foot_vel_treadmill" in this code. It is NOT 
%   recommended to use the velocity readout from the treadmill while 
%   walking. This input should be set to 0 for overground walking.
%
% Returns
% -------
% Dist_Seg_Pow = distal "segment" power
%__________________________________________________________________________
% If treadmill gait, adjust the segment velocity for the backward motion on
% the treadmill belt
if treadmill_yn == 1
	adj_Seg_COM_Vel = Seg_COM_Vel-ones(length(Seg_COM_Vel),1)*foot_vel;
else
    adj_Seg_COM_Vel = Seg_COM_Vel;
end
%__________________________________________________________________________
% Distal Segment power calculations:
Vel_COP_COM = adj_Seg_COM_Vel+cross(Seg_Ang_Vel,(CenterOfPressure-Seg_COM_Pos),2);

Dist_Seg_Pow = dot(GRF,Vel_COP_COM,2)+dot(FreeMoment,Seg_Ang_Vel,2);
end

function foot_vel = find_foot_vel_treadmill(HS,Speed,Foot_Vel)
% This function is used to compute the velocity of the foot on the ground.
% When computing distal "segement" power during walking, it is encouraged
% to compute the treadmill velocity based on foot segment kinematics (or
% foot markers). The treadmill velocity is computed from "foot flat" which 
% is approximately 20% to 40% of stride. This period of "foot flat" was 
% determined from observing a cohort of 10 subjects walking on various 
% speeds and slopes (Honert & Zelik 2019). 
%__________________________________________________________________________
% Parameters
% ----------
% HS = Heel Strike (or foot contact events)
%   Array of heel strikes indicies can be either left or right
% Foot_Vel = velocity of the foot center-of-mass in the lab coordinate
%   system. This is to adjust the velocity of the foot on a treadmill.
% Speed = speed entered into the treadmill (e.g., 1.0)
%   (This variable is for debugging purposes)
%
% Returns
% -------
% foot_vel = average velocity of the foot moving backwards on the treadmill
%__________________________________________________________________________
% Debugging:
debug = 0;  % 1 to compare treadmill velocity and foot velocity
%__________________________________________________________________________
% When using a treadmill when computing distal segment power it is
% necessary to adjust the speed of the treadmill based otherwise there will
% be an artificial power term within the computation. In order to do this,
% it is approximated by the foot_flat variable when the foot is on the
% treadmill and moving at the pace of the treadmill backwards.
%__________________________________________________________________________
% Estimation when foot flat is to adjust for velocity in terms of stride. 
% Examining 20% to 40% of stride has been successful in the past.
foot_flat = [.2,.4];
%__________________________________________________________________________
% Index through the heel strikes 
for ii = 1:(length(HS)-1)
    strideframe = HS(ii+1)-HS(ii);
    vel_foot_flat = (HS(ii)+round(foot_flat(1)*strideframe)):(HS(ii)+round(foot_flat(2)*strideframe));
    stride_foot_velocity(ii,:) = mean(Foot_Vel(vel_foot_flat,:)); 
end
% Debug mode: Show estimated velocity vs. treadmill velocity
if debug == 1
    figure(1010)
    plot(1:length(stride_foot_velocity(:,1)),Speed*ones(1,length(stride_foot_velocity(:,1))),1:length(stride_foot_velocity(:,1)),rssq(stride_foot_velocity,2),'r')
end
foot_vel = mean(stride_foot_velocity,1);
end

function intp_var = intp_stride(HS,var)
% This function interpolates the input variable across stride to 101 points
%__________________________________________________________________________
% Parameters
% ----------
% HS = Heel Strike (foot contact) indices
% var = variable of interest to interpolates
%
% Returns
% -------
% intp_var = interpolated variable to 101 points
%__________________________________________________________________________
% Index through the heel strike indices 
for ii = 1:(length(HS)-1)
	qq = 1:(HS(ii+1) - HS(ii) + 1);
	xx = linspace(1,length(qq),101);
	intp_var(:,ii) = interp1(qq,var(HS(ii):HS(ii+1)),xx,'spline');
end
end

%==========================================================================
